/**
    @file ft51_io_registers.h

    @brief IO register library

    Read, write, set/clear bits of registers for on-chip peripherals.

    @version 1.0.0

**/
/*
 * ============================================================================
 * History
 * =======
 *
 * (C) Copyright 2014, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */

#ifndef __FT51_IO_REGISTERS_H__
#define __FT51_IO_REGISTERS_H__

#include <stdint.h>
#include "ft51.h"

#ifndef IO_REG_GENERAL
#define IO_REG_GENERAL      0
#endif
#ifndef IO_REG_INTERRUPTS
#define IO_REG_INTERRUPTS   1
#endif

/* Read and write */
#define IO_REG_WRITE_CH(channel, address, data) IO_REG_WRITE_CH_(channel, address, data)
#define IO_REG_WRITE_CH_(channel, address, data) \
do { \
    IO_ADDR_##channel##_H = ((uint16_t)(address) >> 8); \
    IO_ADDR_##channel##_L = (uint8_t)(address); \
    IO_DATA_##channel= (data); \
} while (0)

#define IO_REG_READ_CH(channel, address, data) IO_REG_READ_CH_(channel, address, data)
#define IO_REG_READ_CH_(channel, address, data) \
do { \
    IO_ADDR_##channel##_H = ((uint16_t)(address) >> 8); \
    IO_ADDR_##channel##_L = (uint8_t)(address); \
    (data) = IO_DATA_##channel; \
} while (0)

/* Write a 16 bit word across a pair of consecutive 8 bit registers. */
#define IO_REG_WRITE_W_CH(channel, address, data) IO_REG_WRITE_W_CH_(channel, address, data)
#define IO_REG_WRITE_W_CH_(channel, address, data) \
do { \
    IO_ADDR_##channel##_L = (address); \
    IO_DATA_##channel = LSB((data)); \
    IO_ADDR_##channel##_L = (address) + 1; \
    IO_DATA_##channel = MSB((data)); \
} while (0)

// Sets register data bit if corresponding bit in mask is 1
#define IO_REG_SET_MASK_CH(channel, address, mask) IO_REG_SET_MASK_CH_(channel, address, mask)
#define IO_REG_SET_MASK_CH_(channel, address, mask) \
do { \
    uint8_t temp; \
    IO_ADDR_##channel##_H = ((uint16_t)(address) >> 8); \
    IO_ADDR_##channel##_L = (uint8_t)(address); \
    temp = IO_DATA_##channel; \
    IO_DATA_##channel = (temp | (mask)); \
} while (0)

// Clears register data bit if corresponding bit in mask is 1
#define IO_REG_CLEAR_MASK_CH(channel, address, mask) IO_REG_CLEAR_MASK_CH_(channel, address, mask)
#define IO_REG_CLEAR_MASK_CH_(channel, address, mask) \
do { \
    uint8_t clear_mask_temp; \
    IO_ADDR_##channel##_H = ((uint16_t)(address) >> 8); \
    IO_ADDR_##channel##_L = (uint8_t)(address); \
    clear_mask_temp = IO_DATA_##channel; \
    IO_DATA_##channel = (clear_mask_temp & ~(mask)); \
} while (0)


#define IO_REG_GENERAL_WRITE(a, b) IO_REG_WRITE_CH(IO_REG_GENERAL, a, b)
#define IO_REG_GENERAL_READ(a, b) IO_REG_READ_CH(IO_REG_GENERAL, a, b)
#define IO_REG_GENERAL_WRITE_W(a, b) IO_REG_WRITE_W_CH(IO_REG_GENERAL, a, b)
#define IO_REG_GENERAL_SET_MASK(a, b) IO_REG_SET_MASK_CH(IO_REG_GENERAL, a, b)
#define IO_REG_GENERAL_CLEAR_MASK(a, b) IO_REG_CLEAR_MASK_CH(IO_REG_GENERAL, a, b)


#define IO_REG_INTERRUPTS_WRITE(a, b) IO_REG_WRITE_CH(IO_REG_INTERRUPTS, a, b)
#define IO_REG_INTERRUPTS_READ(a, b) IO_REG_READ_CH(IO_REG_INTERRUPTS, a, b)
#define IO_REG_INTERRUPTS_WRITE_W(a, b) IO_REG_WRITE_W_CH(IO_REG_INTERRUPTS, a, b)
#define IO_REG_INTERRUPTS_SET_MASK(a, b) IO_REG_SET_MASK_CH(IO_REG_INTERRUPTS, a, b)
#define IO_REG_INTERRUPTS_CLEAR_MASK(a, b) IO_REG_CLEAR_MASK_CH(IO_REG_INTERRUPTS, a, b)

#endif // __FT51_IO_REGISTERS_H__