/**
    @file spaced_invaders.c

    @brief Sample application to demonstrate an FT51 driving an FT800, 
         with an attempt at reproducing an arcade Space Invaders.

	@details Sample application to demonstrate an FT51 driving an FT800, with 
	an attempt at reproducing an arcade Space Invaders.
	- To fit the code within the 16k flash of the FT51, some of the FT800 
	primitives have been replaced with modified	versions specific to this 
	application. The result is that not only does the function occupy less 
	space, but the function call also uses less, a major saving where a 
	function is used within the code at numerous points.

    @version 1.0.0

**/
/*
 * ============================================================================
 * History
 * =======
 *
 * (C) Copyright 2014, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */

#include "FT_Platform.h"

ft_void_t SpacedInvaders();
static volatile uint8_t  checkButtons = 0; // what switches have been pressed

#define SPI_CLOCK_RATE 12000

// ---------------------------------------------------------------------------------
// Various FT800 specific definitions and variables ...
// ---------------------------------------------------------------------------------
#define SECTION_BEGIN_BITMAPS   BEGIN(BITMAPS)
#define SECTION_BEGIN_POINTS    BEGIN(POINTS)
#define SECTION_BEGIN_LINESTRIP BEGIN(LINE_STRIP)
#define SECTION_CLOSE_BITMAPS   END()
#define SECTION_CLOSE_POINTS    END()
#define SECTION_CLOSE_LINESTRIP END()

// Values specific to WQVGA LCD display (DEFAULT)
#define FT_DISP_WIDTH    480
#define FT_DISP_HEIGHT   272
#define FT_DISP_HCYCLE   548
#define FT_DISP_HOFFSET   43
#define FT_DISP_HSYNC0     0
#define FT_DISP_HSYNC1    41
#define FT_DISP_VCYCLE   292
#define FT_DISP_VOFFSET   12
#define FT_DISP_VSYNC0     0
#define FT_DISP_VSYNC1    10
#define FT_DISP_PCLK       5
#define FT_DISP_SWIZZLE    0
#define FT_DISP_PCLKPOL    1

Ft_Gpu_Hal_Context_t   host;
Ft_Gpu_Hal_Context_t * phost;

// Globals used for building list of display commands ...
ft_uint32_t * pCmdBuffer;
ft_uint32_t   Ft_CmdBuffer_Index = 0;
ft_uint8_t    Ft_CmdBuffer[FT_CMD_FIFO_SIZE];

typedef signed char ft_int8_t;

// ---------------------------------------------------------------------------------
// Millisecond timer related definitions and variables
// ---------------------------------------------------------------------------------
#define MICROSECONDS_TO_TIMER_TICKS(a) (65536 - ((a) * (48 / 12)))
volatile ft_uint16_t ms_timer = 0;

// =================================================================================================================
// Spaced Invaders - #defines
// =================================================================================================================
// game state values ...
#define GAME_START_NEW FALSE
#define GAME_NEXT_WAVE TRUE

// bitmap and cell handles for invaders / explosion ...
#define BH_INVADERSEXPLODE  0
#define CH_INVADERS         0
#define CH_INVADER1         0
#define CH_INVADER2         2
#define CH_INVADER3         4
#define CH_INVADER4         6
#define CH_INVADER5         8
#define CH_EXPLODES        10

// bitmap and cell handles for shields ...
#define BH_SHIELDS  1
#define CH_SHIELD1  0
#define CH_SHIELD2  4
#define CH_SHIELD3  8
#define CH_SHIELD4 12
#define CH_SHIELD5 16

// bitmap and cell handles for missiles ...
#define BH_BULLETS  2
#define CH_BLASTER  0 // player  missile
#define CH_TUMBLER  4 // invader missile 1
#define CH_ZIGZAGS  8 // invader missile 2
#define CH_STINGER 12 // invader missile 3

// bitmap and cell handles for player / saucer ...
#define BH_PLAYERSAUCER 3
#define CH_PLAYER       0
#define CH_SAUCER       2
#define CH_EXPLODEX     4

// bitmap and cell handles for long line ..
#define BH_LONGLINE     4
#define CH_LONGLINE     0

// play area size and offsets to centre on display ...
#define DISPLAY_XPIX 224
#define DISPLAY_YPIX 256
#define DISPLAY_XOFF ((FT_DISP_WIDTH  - DISPLAY_XPIX) / 2)
#define DISPLAY_YOFF ((FT_DISP_HEIGHT - DISPLAY_YPIX) / 2)

// player specific info ...
#define PLAYERS_TOTL 1
#define PLAYERS_XPOS 16
#define PLAYERS_YPOS 8*27
#define PLAYERS_XPIX 15
#define PLAYERS_YPIX  8
#define PLAYERS_XSPD  1
#define PLAYERS_YSPD  0
#define PLAYERS_XMIN  4
#define PLAYERS_XMAX DISPLAY_XPIX - 20

// shield specific info ...
#define SHIELDS_TOTL 4*10
#define SHIELDS_XPOS 28
#define SHIELDS_YPOS 8*24
#define SHIELDS_XPIX 6
#define SHIELDS_YPIX 6
#define SHIELDS_XGRD (SHIELDS_XPIX * 4) + 24
#define SHIELDS_YGRD 0

// invader specific info ...
#define INVADER_COLS 11
#define INVADER_ROWS  5
#define INVADER_TOTL INVADER_COLS * INVADER_ROWS

#define INVADER_XMIN 8
#define INVADER_XMAX DISPLAY_XPIX - 12 - 8
#define INVADER_YMIN 8*8
#define INVADER_YMAX PLAYERS_YPOS

#define INVADER_XPIX 11
#define INVADER_YPIX  8
#define INVADER_XGRD 16
#define INVADER_YGRD 16
#define INVADER_XSPD  2
#define INVADER_YSPD  8
#define INVADER_LEFT  0
#define INVADER_RIGHT 1

// bullet specific info ...
#define BULLETS_XPIX 3
#define BULLETS_YPIX 8
#define BULLETS_YMIN 8*4
#define BULLETS_YMAX 8*29
#define BULLETS_YSPD 2
#define BLASTER_YSPD 8

// saucers specific info ...
#define SAUCERS_XPIX 16
#define SAUCERS_YPIX  8
#define SAUCERS_XMIN 0
#define SAUCERS_XMAX DISPLAY_XPIX - SAUCERS_XPIX
#define SAUCERS_YMIN 32
#define SAUCERS_YMAX 32
#define SAUCERS_XPOS 16
#define SAUCERS_YPOS 8*5

// default refresh times for update of objects (in milliseconds) ...
#define TIMER_OFF                   0
#define TIMER_NOW                   1
#define TIMER_DISPLAY              20
#define TIMER_PLAYERS              10
#define TIMER_SAUCERS              40
#define TIMER_SAUCERS_NEXT       5000
#define TIMER_INVADER              10
#define TIMER_SHIELDS               0
#define TIMER_BLASTER              20
#define TIMER_MISSILE_DROPPING     20
#define TIMER_MISSILE_WAVE_START   (60 * TIMER_MISSILE_DROPPING)
#define TIMER_MISSILE_UPTO_0200    (48 * TIMER_MISSILE_DROPPING)
#define TIMER_MISSILE_UPTO_1000    (16 * TIMER_MISSILE_DROPPING)
#define TIMER_MISSILE_UPTO_2000    (11 * TIMER_MISSILE_DROPPING)
#define TIMER_MISSILE_UPTO_3000    ( 8 * TIMER_MISSILE_DROPPING)
#define TIMER_MISSILE_OVER_3000    ( 7 * TIMER_MISSILE_DROPPING)
#define TIMER_BUTTONS              10
#define TIMER_EXPLODE             200
#define TIMER_SAUCER_SCORE        500
#define TIMER_SPAWN               100

// number of animation frames for each object ...
#define FRAME_PLAYER  2
#define FRAME_SAUCER  2
#define FRAME_INVADER 2
#define FRAME_BLASTER 4
#define FRAME_MISSILE 4

// buttons for control ...
#define BUTTON_NONE  0x00
#define BUTTON_LEFT  0x01
#define BUTTON_RIGHT 0x02
#define BUTTON_START 0x08
#define BUTTON_FIRE  0x08
#define BUTTON_MASK  (BUTTON_LEFT | BUTTON_RIGHT | BUTTON_START | BUTTON_FIRE)

// =================================================================================================================
// Spaced Invaders - Enumerations, structures, tables etc.
// =================================================================================================================
// life states (used by all objects) ...
typedef enum
{
	LIFE_SPAWN = 0, // about to live
	LIFE_KAPUT,     // dead
	LIFE_TOAST,     // dying
	LIFE_ALIVE,     // 1 hit  left
	LIFE_ALIVE2,    // 2 hits left
	LIFE_ALIVE3,    // 3 hits left
	LIFE_SHIELD,    // 4 hits left
	LIFE_TOTAL
} eLIFESTATE;

// display refresh types (used for update of objects)
typedef enum
{
	// Must be the first three entries (used as array index)...
	REFRESH_TUMBLER = 0,
	REFRESH_ZIGZAGS,
	REFRESH_STINGER,
	// The rest can be in any order ...
	REFRESH_DISPLAY,
	REFRESH_SAUCERS,
	REFRESH_INVADER,
	REFRESH_SHIELDS,
	REFRESH_PLAYERS,
	REFRESH_BLASTER,
	REFRESH_BUTTONS,
	REFRESH_MISSILE,
	// Must be last entry ...
	REFRESH_TOTAL
} eREFRESH;

typedef struct SAMAPP_Bitmap_header
{
	ft_uint8_t Format;
	ft_uint8_t Width;
	ft_uint8_t Height;
	ft_uint8_t Stride;
	ft_int32_t Offset;
} SAMAPP_Bitmap_header_t;

typedef struct siObject
{
	ft_uint8_t xcrd;
	ft_uint8_t ycrd;
	ft_uint8_t type;
	ft_uint8_t life;
	ft_uint8_t anim;
} siObject_t;

typedef struct siText
{
	ft_uint8_t   xoff;
	ft_uint8_t   yoff;
	ft_uint8_t * text;
} siText_t;

// Header of raw data containing properties of bitmap
__code const SAMAPP_Bitmap_header_t InvadersExplodeImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1, 12, 96, 2, 0 }
};

__code const ft_prog_uchar8_t InvadersExplodeImageData[192] =
{
	// 'resolution ', 12, 'x', 96, 'format ', 'L1', 'stride ', 2, ' total size ', 192
	  6,  0, 15,  0, 31,128, 54,192, 63,192, 22,128, 32, 64, 16,128, // invader 1-1
	  6,  0, 15,  0, 31,128, 54,192, 63,192,  9,  0, 22,128, 41, 64, // invader 1-2
	 16, 64,  8,128, 31,192, 59, 96,127,240, 95,208, 80, 80, 13,128, // invader 2-1
	 16, 64, 72,144, 95,208,118,240, 63,224, 31,192, 16, 64, 32, 32, // invader 2-2
	 16, 64,  8,128, 31,192, 59, 96,127,240, 95,208, 80, 80, 13,128, // invader 3-1
	 16, 64, 72,144, 95,208,118,240, 63,224, 31,192, 16, 64, 32, 32, // invader 3-2
	 15,  0,127,224,255,240,206,112,255,240, 57,192,102, 96, 48,192, // invader 4-1
	 15,  0,127,224,255,240,231, 48,255,240, 25,128, 54,192,192, 48, // invader 4-2
	 15,  0,127,224,255,240,206,112,255,240, 57,192,102, 96, 48,192, // invader 5-1
	 15,  0,127,224,255,240,231, 48,255,240, 25,128, 54,192,192, 48, // invader 5-2
	  0,  0, 16, 32,  1,  0, 64,  0,  2, 32,  8,  0,  0,128,  0,  0, // explode 1-1
	  8, 64,200,128, 32,128,  4, 48,194,  0, 16, 64, 17, 48, 33,  0  // explode 1-2
};

// Header of raw data containing properties of bitmap
__code const SAMAPP_Bitmap_header_t ShieldsImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1, 6, 120, 1, 0 }
};

__code const ft_prog_uchar8_t ShieldsImageData[120] =
{
	// 'resolution ', 6, 'x', 120, 'format ', 'L1', 'stride ', 1, ' total size ', 120
	 28, 60,124,252,252,252, // shield 1-1
	 28, 56, 44,124,204,244, // shield 1-2
	 20, 24, 32,124,136,148, // shield 1-3
	  0,  8, 32, 36,136,  0, // shield 1-4
	224,240,248,252,252,252, // shield 2-1
	128,240, 40,124,204,244, // shield 2-2
	  0, 80, 32,124,136,148, // shield 2-3
	  0, 64, 32, 36,136,  0, // shield 2-4
	252,252,240,224,192,128, // shield 3-1
	156,248, 32, 96,192,128, // shield 3-2
	 20, 88, 32, 96,128,128, // shield 3-3
	  0, 72, 32, 32,128,  0, // shield 3-4
	252,252, 60, 28, 12,  4, // shield 4-1
	156,248, 44, 28, 12,  4, // shield 4-2
	 20, 88, 32, 28,  8,  4, // shield 4-3
	  0, 72, 32,  4,  8,  0, // shield 4-4
	252,252,252,252,252,252, // shield 5-1
	156,248, 44,124,204,244, // shield 5-2
	 20, 88, 32,124,136,148, // shield 5-3
	  0, 72, 32, 36,136,  0, // shield 5-4
};

// Header of raw data containing properties of bitmap
__code const SAMAPP_Bitmap_header_t MissileImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1, 3, 128, 1, 0 }
};

__code const ft_prog_uchar8_t MissileImageData[128] =
{
	// 'resolution ', 3, 'x', 128, 'format ', 'L1', 'stride ', 1, ' total size ', 128
	 64, 64, 64, 64,  0,  0,  0,  0, // blaster 1-1
	 64, 64, 64, 64,  0,  0,  0,  0, // blaster 1-2
	 64, 64, 64, 64,  0,  0,  0,  0, // blaster 1-3
	 64, 64, 64, 64,  0,  0,  0,  0, // blaster 1-4
	224, 64, 64, 64, 64, 64, 64, 64, // tumbler 1-1
	 64, 64,224, 64, 64, 64, 64, 64, // tumbler 1-2
	 64, 64, 64, 64, 64,224, 64, 64, // tumbler 1-3
	 64, 64, 64, 64, 64, 64, 64,224, // tumbler 1-4
	 64, 32, 64,128, 64, 32, 64,128, // zigzags 1-1
	 32, 64,128, 64, 32, 64,128, 64, // zigzags 1-2
	 64,128, 64, 32, 64,128, 64, 32, // zigzags 1-3
	128, 64, 32, 64,128, 64, 32, 64, // zigzags 1-4
	 64, 32, 64,128, 64, 64, 64, 64, // stinger 1-1
	 64, 64, 64, 64, 32, 64,128, 64, // stinger 1-2
	 64, 64, 32, 64,128, 64, 64, 64, // stinger 1-3
	 64, 64, 64, 32, 64,128, 64, 64  // stinger 1-4
};

// Header of raw data containing properties of bitmap
__code const SAMAPP_Bitmap_header_t PlayerSaucerImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1, 16, 40, 2, 0 }
};

__code const ft_prog_uchar8_t PlayerSaucerImageData[80] =
{
	// 'resolution ', 16, 'x', 40, 'format ', 'L1', 'stride ', 2, ' total size ', 80
	  1,  0,  3,128,  3,128,127,252,255,254,255,254,255,254,255,254, // player base 1-1
	  0,  0,  2,128,  2,128,126,252,255,254,255,254,255,254,255,254, // player base 1-2
	  0,  0,  7,224, 31,248, 63,252,109,182,255,255, 57,156, 16,  8, // saucer 1-1
	  0,  0,  7,224, 31,248, 63,252,114, 78,255,255, 57,156, 16,  8, // saucer 1-2
	  0,  0, 36, 36, 18, 72,  8, 16, 96,  6,  8, 16, 18, 72, 36, 36  // explode X
};

// Header of raw data containing properties of bitmap
__code const SAMAPP_Bitmap_header_t LongLineImageInfo[1] =
{
	// format,width,height,stride,arrayoffset
	{ L1, 224, 1, 28, 0 }
};

__code const ft_prog_uchar8_t LongLineImageData[28] =
{
	// 'resolution ', 224, 'x', 1, 'format ', 'L1', 'stride ', 28, ' total size ', 28
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255
};

__code const ft_uint8_t invaderScore[5] =
{
	4, 2, 2, 1, 1
};

// Count of player bullets kept which wraps around after 15 shots and is used to
// determine score for saucer, so to score 300 each time shoot 1st saucer with
// 23rd bullet and then every 15th bullet after that ...
__code const ft_uint8_t saucersScore[15] =
{
//   1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
	10, 5,10,15,10,10, 5,30,10,10,10, 5,15,10, 5
};

// Text for displaying score for saucer hit ...
__code const ft_uint8_t * textForScore[6] =
{
	"50", "100", "150", "?", "?", "300"
};

__code const ft_int32_t colours[24] =
{
	COLOR_RGB( 255,  63,   0 ),
	COLOR_RGB( 255, 127,   0 ), //  1
	COLOR_RGB( 255, 191,   0 ),
	COLOR_RGB( 255, 255,   0 ), //  2
	COLOR_RGB( 191, 255,   0 ),
	COLOR_RGB( 127, 255,   0 ), //  3
	COLOR_RGB(  63, 255,   0 ),
	COLOR_RGB(   0, 255,   0 ), //  4
	COLOR_RGB(   0, 255,  63 ),
	COLOR_RGB(   0, 255, 127 ), //  5
	COLOR_RGB(   0, 255, 191 ),
	COLOR_RGB(   0, 255, 255 ), //  6
	COLOR_RGB(   0, 191, 255 ),
	COLOR_RGB(   0, 127, 255 ), //  7
	COLOR_RGB(   0,  63, 255 ),
	COLOR_RGB(   0,   0, 255 ), //  8
	COLOR_RGB(  63,   0, 255 ),
	COLOR_RGB( 127,   0, 255 ), //  9
	COLOR_RGB( 191,   0, 255 ),
	COLOR_RGB( 255,   0, 255 ), // 10
	COLOR_RGB( 255,   0, 191 ),
	COLOR_RGB( 255,   0, 127 ), // 11
	COLOR_RGB( 255,   0,  63 ),
	COLOR_RGB( 255,   0,   0 )  // 12
};

// Data containing x,y positions and cell types that make up a shield ...
__code const ft_uint8_t shieldXcrdData[10] =
{
	 0,	 6,	12,	18, // top row
	 0,	 6,	12,	18, // middle row
	 0,         18, // bottom row
};

__code const ft_uint8_t shieldYcrdData[10] =
{
	 0,  0,  0,  0, // top row
	 6,  6,  6,  6, // middle row
	12,         12  // bottom row
};

__code const ft_uint8_t shieldBitsData[10] =
{
	CH_SHIELD1,	CH_SHIELD5, CH_SHIELD5, CH_SHIELD2, // top row
	CH_SHIELD5, CH_SHIELD3, CH_SHIELD4, CH_SHIELD5, // middle row
	CH_SHIELD5,                         CH_SHIELD5  // bottom row
};

// invader columns to drop missiles in ...
__code const ft_uint8_t dropsTumbler[] = { 1, 7, 1, 1, 1, 4,11, 1, 6, 3, 1, 1,11, 9, 2, 8, 0};
__code const ft_uint8_t dropsZigZags[] = {11, 1, 6, 3, 1, 1,11, 9, 2, 8, 2,11, 4, 7,10, 0};

// invader wave start height according to number of waves completed ...
__code const ft_uint8_t invaderWave[] = { 1, 2, 3, 3, 3, 4, 4, 4};

// =================================================================================================================
// Spaced Invaders - Global variables
// =================================================================================================================
siObject_t sigodInvader[INVADER_TOTL];	// invader info
siObject_t sigodShields[SHIELDS_TOTL];	// shields info
siObject_t sigodPlayers;				// player  info
siObject_t sigodSaucers;				// saucer  info
siObject_t sigodBlaster;				// missile info (player)
siObject_t sigodMissile[3];				// missile info (invader TUMBLER, ZIGZAGS and STINGER)

// need to keep this separate from the text list below as we write the score directly into it
ft_uchar8_t displayScore[] =  "000000   000000   000000";  // line  1, offset  2

__code static const siText_t siTextList[10] =
{
	{ 1, 1, "SCORE<1> HI-SCORE SCORE<2>"}, // offset  1, line  1
	//       "000000   000000   000000"
	{ 2, 3, displayScore                }, // offset  2, line  3
	{12, 8,            "PLAY"           }, // offset 12, line  8
	{ 7,11,       "SPACE  INVADERS"     }, // offset  7, line 11
	{ 4,15,    "*SCORE ADVANCE TABLE*"  }, // offset  4, line 15
	{10,17,          "=? MYSTERY"       }, // offset 10, line 17
	{10,19,          "=30 POINTS"       }, // offset 10, line 19
	{10,21,          "=20 POINTS"       }, // offset 10, line 21
	{10,23,          "=10 POINTS"       }, // offset 10, line 23
	{17,30,                 "CREDIT 01" }  // offset 17, line 30
};

ft_bool_t  invaderEdges; // flag to mark that an invader has reached left/right edge of display
ft_bool_t  invaderLands; // flag to mark that an invader has reached bottom of display
ft_uint8_t invaderAlive; // current number of invaders still alive
ft_uint8_t invaderStart; // Y position to start invader wave at
ft_uint8_t invaderMoveX; // used to move invaders backwards and forwards
ft_uint8_t invaderMoveY; // used to move invaders downwards
ft_uint8_t invaderCheck; // next invader to check (move it if alive)

ft_uint8_t playersXcord;
ft_uint8_t playersYcord;
ft_uint8_t playersMoveX;
ft_uint8_t playersLives;
ft_uint8_t playersSpawn; // Used to flash player on/off during spawning
ft_uint8_t blasterCount; // Current number of player shots fired
ft_int8_t  saucersMoveX;

// counters for update of objects ...
volatile ft_uint16_t updateTimer[REFRESH_TOTAL];

// scores ...
ft_int16_t   scoreCurrent = 0;
ft_int16_t   scoreHighest = 0;
ft_uint8_t * scoreSaucer;

// guess ...
ft_uint8_t currentWave;

// missile control indexing ...
ft_uint8_t indexZigZags;
ft_uint8_t indexTumbler;

// addresses of bitmap data to be stored in FT800 graphics ram ...
ft_int32_t ramInvadersExplode;
ft_int32_t ramShields;
ft_int32_t ramMissile;
ft_int32_t ramPlayerSaucer;
ft_int32_t ramLongLine;
ft_int32_t ramNextBit;

// ---------------------------------------------------------------------------------
// FT800 routines (modified versions of library code to reduce code size)
// ---------------------------------------------------------------------------------
// Modified version of <<<  >>>
ft_void_t FT800_Write08(ft_uint32_t addr, ft_uint8_t data)
{
	Ft_Gpu_Hal_StartTransfer(phost, FT_GPU_WRITE, addr);
	Ft_Gpu_Hal_Transfer8(phost, data);
	Ft_Gpu_Hal_EndTransfer(phost);
}

// Modified version of <<<  >>>
ft_void_t FT800_Write16(ft_uint32_t addr, ft_uint16_t data)
{
	Ft_Gpu_Hal_StartTransfer(phost, FT_GPU_WRITE, addr);
	Ft_Gpu_Hal_Transfer16(phost, data);
	Ft_Gpu_Hal_EndTransfer(phost);
}

// Modified version of <<<  >>>
ft_void_t FT800_CmdBuffer_AddCmd(ft_uint32_t cmd)
{
	// Copy the command instruction into buffer
	pCmdBuffer  = (ft_uint32_t *)&Ft_CmdBuffer[Ft_CmdBuffer_Index];
	*pCmdBuffer = cmd;
	// Increment the command index
	Ft_CmdBuffer_Index += FT_CMD_SIZE;
}

// Modified version of <<< Ft_Gpu_Hal_WrCmdBuf(...) >>>
ft_void_t FT800_CmdBuffer_Send()
{
	if (Ft_CmdBuffer_Index > 0)
	{
		ft_uint8_t * databuff = Ft_CmdBuffer;
		ft_uint16_t  datasize = Ft_CmdBuffer_Index;

		Ft_Gpu_Hal_CheckCmdBuffer(phost, datasize);
		Ft_Gpu_Hal_StartCmdTransfer(phost, FT_GPU_WRITE, datasize);

		while (datasize--)
		{
			Ft_Gpu_Hal_Transfer8(phost, *databuff);
			databuff++;
		}

		Ft_Gpu_Hal_EndTransfer(phost);
		Ft_Gpu_Hal_Updatecmdfifo(phost, Ft_CmdBuffer_Index);
		Ft_Gpu_Hal_WaitCmdfifo_empty(phost);

		Ft_CmdBuffer_Index = 0;
	}
}

// ---------------------------------------------------------------------------------
// Millisecond timer ISR
// ---------------------------------------------------------------------------------
void ms_timer_interrupt(const uint8_t flags)
{
	ft_uint8_t timer;

    (void) flags; // Flags not currently used

    if (ms_timer)
    {
        ms_timer--;
    }

    // refresh all update timers ...
    for (timer = 0; timer < REFRESH_TOTAL; timer++)
    {
        if (updateTimer[timer])
        {
        	updateTimer[timer]--;
        }
    }

    // Reload the timer
    TH0 = MSB(MICROSECONDS_TO_TIMER_TICKS(1000));
    TL0 = LSB(MICROSECONDS_TO_TIMER_TICKS(1000));
}

// ---------------------------------------------------------------------------------
// Millisecond timer interrupt initialisation
// ---------------------------------------------------------------------------------
void ms_timer_initialise(void)
{
    // Register our own handler for interrupts from Timer 0
    interrupts_register(ms_timer_interrupt, interrupts_timer0);

    // Timer0 is controlled by TMOD bits 0 to 3, and TCON bits 4 to 5.
    TMOD &= 0xF0; // Clear Timer0 bits
    TMOD |= 0x01; // Put Timer0 in mode 1 (16 bit)

    // Set the count-up value so that it rolls over to 0 after 1 millisecond.
    TH0 = MSB(MICROSECONDS_TO_TIMER_TICKS(1000));
    TL0 = LSB(MICROSECONDS_TO_TIMER_TICKS(1000));

    TCON &= 0xCF; // Clear Timer0's Overflow and Run flags
    TCON |= 0x10; // Start Timer0 (set its Run flag)
}

// ---------------------------------------------------------------------------------
// Boot configuration
// ---------------------------------------------------------------------------------
ft_void_t Ft_BootupConfig()
{
    // Boot up for FT800 followed by graphics primitive sample cases
    // Initial boot up DL - make the back ground green color
	ft_uint8_t FT_DLCODE_BOOTUP[12] =
    {
      240,240,240,  2, // GPU instruction CLEAR_COLOR_RGB
        7,  0,  0, 38, // GPU instruction CLEAR
        0,  0,  0,  0, // GPU instruction DISPLAY
    };

	// Do a power cycle
	Ft_Gpu_Hal_Powercycle(phost,FT_TRUE);
	Ft_Gpu_Hal_Rd16(phost,RAM_G);
	// Set the clk to external clock
	Ft_Gpu_HostCommand(phost,FT_GPU_EXTERNAL_OSC);
	Ft_Gpu_Hal_Sleep(10);

	// Switch PLL output to 48MHz
	Ft_Gpu_HostCommand(phost,FT_GPU_PLL_48M);
	Ft_Gpu_Hal_Sleep(10);

	// Do a core reset for safer side
	Ft_Gpu_HostCommand(phost,FT_GPU_CORE_RESET);

	// Access address 0 to wake up the FT800
	Ft_Gpu_HostCommand(phost,FT_GPU_ACTIVE_M);

	FT800_Write08(REG_GPIO_DIR,0x80 | Ft_Gpu_Hal_Rd8(phost,REG_GPIO_DIR));
	FT800_Write08(REG_GPIO,    0x80 | Ft_Gpu_Hal_Rd8(phost,REG_GPIO));

	{
		ft_uint8_t chipid;
		// Read Register ID to check if FT800 is ready.
		chipid = Ft_Gpu_Hal_Rd8(phost, REG_ID);
		while(chipid != 0x7C)
			chipid = Ft_Gpu_Hal_Rd8(phost, REG_ID);
	}

	// Configuration of LCD display
	FT800_Write16(REG_HCYCLE,   FT_DISP_HCYCLE);
	FT800_Write16(REG_HOFFSET,  FT_DISP_HOFFSET);
	FT800_Write16(REG_HSYNC0,   FT_DISP_HSYNC0);
	FT800_Write16(REG_HSYNC1,   FT_DISP_HSYNC1);
	FT800_Write16(REG_VCYCLE,   FT_DISP_VCYCLE);
	FT800_Write16(REG_VOFFSET,  FT_DISP_VOFFSET);
	FT800_Write16(REG_VSYNC0,   FT_DISP_VSYNC0);
	FT800_Write16(REG_VSYNC1,   FT_DISP_VSYNC1);
	FT800_Write08(REG_SWIZZLE,  FT_DISP_SWIZZLE);
	FT800_Write08(REG_PCLK,     FT_DISP_PCLK); // after this display is visible on the LCD
	FT800_Write08(REG_PCLK_POL, FT_DISP_PCLKPOL);
	FT800_Write16(REG_HSIZE,    FT_DISP_WIDTH);
	FT800_Write16(REG_VSIZE,    FT_DISP_HEIGHT);

	// Touch configuration - configure the resistance value to 1200 - this value is specific to customer requirement and derived by experiment
	FT800_Write16(REG_TOUCH_RZTHRESH,1200);
	FT800_Write08(REG_GPIO_DIR,      0xff);
	FT800_Write08(REG_GPIO,          0xff);

	//It is optional to clear the screen here
	Ft_Gpu_Hal_WrMem(phost, RAM_DL,(ft_uint8_t *)FT_DLCODE_BOOTUP,sizeof(FT_DLCODE_BOOTUP));
	FT800_Write08(REG_DLSWAP,DLSWAP_FRAME);
}

ft_void_t setup()
{
	Ft_Gpu_HalInit_t halinit;
	halinit.TotalChannelNum = 1;

	Ft_Gpu_Hal_Init(&halinit);
	host.hal_config.channel_no = 1;
	host.hal_config.spi_clockrate_khz = SPI_CLOCK_RATE;
    Ft_Gpu_Hal_Open(&host);

	phost = &host;

    Ft_BootupConfig();
    //Show the booting up screen.
    Ft_Gpu_Hal_Sleep(2000);

    // Setup switches for FT51 ...
    IOMUX_INPUT(DIO_0, IOMUX_IN_PORT0_1); // SW1 = bit 0 = LEFT
    IOMUX_INPUT(DIO_7, IOMUX_IN_PORT0_2); // SW2 = bit 1 = RIGHT
    IOMUX_INPUT(AIO_7, IOMUX_IN_PORT0_3); // SW3 = bit 2 = N/A
    IOMUX_INPUT(AIO_8, IOMUX_IN_PORT0_4); // SW4 = bit 3 = START/FIRE

    SpacedInvaders();

	// Close all the opened handles
    Ft_Gpu_Hal_Close(phost);
    Ft_Gpu_Hal_DeInit();
}

void loop()
{
}

// =================================================================================================================
// SPACED-INVADERS - Object setup and VertexII function
// =================================================================================================================
ft_void_t SI_SetupObject(siObject_t * data, ft_uint8_t xcrd, ft_uint8_t ycrd, ft_uint8_t type, ft_uint8_t life)
{
	data->xcrd = xcrd;
	data->ycrd = ycrd;
	data->type = type;
	data->life = life;
	data->anim = 0;
}

// about 76 bytes per call of this function as opposed to 176 when using FT800_CmdBuffer_AddCmd(VERTEX2II(x,y,h,c))
ft_void_t SI_VertexII(ft_uint16_t xcrd, ft_uint16_t ycrd, ft_uchar8_t hndl, ft_uchar8_t cell)
{
	FT800_CmdBuffer_AddCmd(VERTEX2II(DISPLAY_XOFF + xcrd, DISPLAY_YOFF + ycrd, hndl, cell));
}

// =================================================================================================================
// SPACED-INVADERS - Initialisation
// =================================================================================================================
ft_void_t SI_BeginGame(ft_bool_t gamestate)
{
	siObject_t * data;
	ft_uint8_t   rows;
	ft_uint8_t   cols;
	ft_uint8_t   xcrd;
	ft_uint8_t   ycrd;
	//ft_uint8_t   cell;

	if (gamestate == GAME_START_NEW)
	{
		playersLives = 3;
		currentWave  = 0;
		blasterCount = 0;
		scoreCurrent = 0;
		saucersMoveX = 2;
		indexTumbler = 0;
		indexZigZags = 0;
		SI_SetupObject(&sigodPlayers, PLAYERS_XPOS, PLAYERS_YPOS, CH_PLAYER, LIFE_ALIVE);
	}
	else
	{
		currentWave = ++currentWave % 8;
	}

	// determine vertical start position of invaders according to wave ...
	invaderStart = INVADER_YMIN + invaderWave[currentWave] * INVADER_YSPD;

	// Initialise saucer ...
	SI_SetupObject(&sigodSaucers, 0, 0, 0, LIFE_KAPUT);

	// Initialise invaders ...
	invaderAlive = INVADER_TOTL;
	invaderMoveX = INVADER_RIGHT;
	invaderMoveY = 0;
	invaderEdges = FALSE; // no invaders have reached edges  of display
	invaderLands = FALSE; // no invaders have reached bottom of display
	invaderCheck = 0;     // next invader to update

	data = sigodInvader;
	ycrd = invaderStart + 7 * 8;
	for (rows = 0; rows < INVADER_ROWS; rows++)
	{
		xcrd = INVADER_XMIN;
		for (cols = 0; cols < INVADER_COLS; cols++)
		{
			SI_SetupObject(data, xcrd, ycrd, CH_INVADERS + (8 - (rows * 2)), LIFE_SPAWN);
			xcrd += INVADER_XGRD;
			data++;
		}
		ycrd -= INVADER_YGRD;
	}

	// Initialise shields ...
	data = sigodShields;
	xcrd = SHIELDS_XPOS;
	for (cols = 0; cols < 4; cols++)
	{
		for (rows = 0; rows < 10; rows++)
		{
			SI_SetupObject(data, xcrd + shieldXcrdData[rows], SHIELDS_YPOS + shieldYcrdData[rows], shieldBitsData[rows], LIFE_SHIELD);
			data++;
		}
		xcrd += SHIELDS_XGRD;
	}

	// Initialise player blaster and invader missiles ...
	SI_SetupObject(&sigodBlaster,                  0, 0, CH_BLASTER, LIFE_KAPUT);
	SI_SetupObject(&sigodMissile[REFRESH_TUMBLER], 0, 0, CH_TUMBLER, LIFE_KAPUT);
	SI_SetupObject(&sigodMissile[REFRESH_ZIGZAGS], 0, 0, CH_ZIGZAGS, LIFE_KAPUT);
	SI_SetupObject(&sigodMissile[REFRESH_STINGER], 0, 0, CH_STINGER, LIFE_KAPUT);

	// Initialise timers for movement of objects  ...
	updateTimer[REFRESH_DISPLAY] = TIMER_NOW;
	updateTimer[REFRESH_SAUCERS] = TIMER_SAUCERS_NEXT;
	updateTimer[REFRESH_INVADER] = TIMER_NOW;
	updateTimer[REFRESH_SHIELDS] = TIMER_NOW;
	updateTimer[REFRESH_PLAYERS] = TIMER_NOW;
	updateTimer[REFRESH_BLASTER] = TIMER_OFF;
	updateTimer[REFRESH_TUMBLER] = TIMER_OFF;
	updateTimer[REFRESH_ZIGZAGS] = TIMER_OFF;
	updateTimer[REFRESH_STINGER] = TIMER_OFF;
	updateTimer[REFRESH_BUTTONS] = TIMER_NOW;
	// Timer REFRESH_MISSILE is used to determine when first missile is fired after the start of a wave,
	// and subsequently the period between missiles being fired (which decreases as the score increases).
	updateTimer[REFRESH_MISSILE] = TIMER_MISSILE_WAVE_START;
}

// =================================================================================================================
// SPACED-INVADERS - Object collision
// =================================================================================================================
ft_uint8_t SI_BulletCollide(ft_uint8_t bx, ft_uint8_t by, siObject_t * item, ft_uint8_t cw, ft_uint8_t ch)
{
	ft_uint8_t state = LIFE_ALIVE;
	ft_uint8_t xmin;
	ft_uint8_t xmax;
	ft_uint8_t ymin;
	ft_uint8_t ymax;

	if (item->life >= LIFE_ALIVE)
	{
		xmin = item->xcrd;
		xmax = xmin + cw;

		if ((bx >= xmin) && (bx <= xmax))
		{
			ymin = item->ycrd;
			ymax = ymin + ch;

			if ((by >= ymin) && (by <= ymax))
			{
				// kill bullet ...
				state = LIFE_TOAST;
				// decrement life of object hit (shields take multiple hits) ...
				item->life--;
			}
		}
	}

	return state;
}

ft_uint8_t SI_HitsOnPlayers(ft_uint8_t bx, ft_uint8_t by)
{
	return SI_BulletCollide(bx, by, &sigodPlayers, PLAYERS_XPIX, PLAYERS_YPIX);
}

ft_uint8_t SI_HitsOnShields(ft_uint8_t bx, ft_uint8_t by)
{
	ft_uint8_t index = 0;
	ft_uint8_t state = LIFE_ALIVE;

	do
	{
		state = SI_BulletCollide(bx, by, &sigodShields[index], SHIELDS_XPIX, SHIELDS_YPIX);
	} while ((state == LIFE_ALIVE) && (++index < SHIELDS_TOTL));

	return state;
}

ft_uint8_t SI_HitsOnInvader(ft_uint8_t bx, ft_uint8_t by)
{
	ft_uint8_t index = 0;
	ft_uint8_t state = LIFE_ALIVE;

	do
	{
		state = SI_BulletCollide(bx, by, &sigodInvader[index], INVADER_XPIX, INVADER_YPIX);

		if (state == LIFE_TOAST)
		{
			// decrement number of invaders still alive ...
			invaderAlive--;

			// update current score according to type of invader ...
			scoreCurrent += invaderScore[ sigodInvader[index].type >> 1 ];

			// invader has been hit so will pause for a brief time ...
			updateTimer[REFRESH_INVADER] = TIMER_EXPLODE;
		}
	} while ((state == LIFE_ALIVE) && (++index < INVADER_TOTL));

	return state;
}

ft_uint8_t SI_HitsOnSaucer(ft_uint8_t bx, ft_uint8_t by)
{
	ft_uint8_t state;

	state = SI_BulletCollide(bx, by, &sigodSaucers, SAUCERS_XPIX, SAUCERS_YPIX);

	if (state == LIFE_TOAST)
	{
		// update current score according to number of blaster shots fired ...
		scoreCurrent += saucersScore[blasterCount];

		// make a note of the score and display it for a while ...
		scoreSaucer   = (ft_uint8_t *)textForScore[ (saucersScore[blasterCount] / 5) - 1 ];
		updateTimer[REFRESH_SAUCERS] = TIMER_SAUCER_SCORE;
	}

	return state;
}

ft_uint8_t SI_HitsOnTumbler(ft_uint8_t bx, ft_uint8_t by)
{
	return SI_BulletCollide(bx, by, &sigodMissile[REFRESH_TUMBLER], BULLETS_XPIX, BULLETS_YPIX);
}

ft_uint8_t SI_HitsOnZigZags(ft_uint8_t bx, ft_uint8_t by)
{
	return SI_BulletCollide(bx, by, &sigodMissile[REFRESH_ZIGZAGS], BULLETS_XPIX, BULLETS_YPIX);
}

ft_uint8_t SI_HitsOnStinger(ft_uint8_t bx, ft_uint8_t by)
{
	return SI_BulletCollide(bx, by, &sigodMissile[REFRESH_STINGER], BULLETS_XPIX, BULLETS_YPIX);
}

// =================================================================================================================
// SPACED-INVADERS - Update object positions
// =================================================================================================================
#define SI_TIMER_EXPIRED(timer) updateTimer[timer] == 0

ft_uint8_t InvaderColumn(ft_uint8_t column)
{
	ft_uint8_t find = 0;
	column -= 1;

	// check to see if an invader is alive in the column specified ...
	for (find = 0; find < 5; find++)
	{
		if (sigodInvader[column].life == LIFE_ALIVE)
		{
			return column;
		}
		column += INVADER_COLS;
	}

	return 255;
}

ft_uint8_t InvaderNearest()
{
	ft_uint8_t index;

	// find the nearest invader vertically above the player (within 4 pixels of centre) ...
	for (index = 0; index < INVADER_TOTL; index++)
	{
		if (sigodInvader[index].life == LIFE_ALIVE)
		{
			if ( abs(sigodInvader[index].xcrd - sigodPlayers.xcrd) < 4)
			{
				return index;
			}
		}
	}

	return 255;
}

ft_void_t SI_UpdateSaucers()
{
	if (SI_TIMER_EXPIRED(REFRESH_SAUCERS))
	{
		switch (sigodSaucers.life)
		{
		case LIFE_ALIVE:
			// update position and frame ...
			sigodSaucers.xcrd += saucersMoveX;
			sigodSaucers.anim  = ++sigodSaucers.anim % FRAME_SAUCER;

			if ( (sigodSaucers.xcrd <= SAUCERS_XMIN) || (sigodSaucers.xcrd >= SAUCERS_XMAX) )
			{
				sigodSaucers.life            = LIFE_KAPUT;         // kill off saucer if it has reached edge of display
				updateTimer[REFRESH_SAUCERS] = TIMER_SAUCERS_NEXT; // delay before next update
			}
			else
			{
				updateTimer[REFRESH_SAUCERS] = TIMER_SAUCERS; // delay before next update
			}
			break;

		case LIFE_TOAST:
			sigodSaucers.life = LIFE_KAPUT;
			updateTimer[REFRESH_SAUCERS] = TIMER_SAUCERS_NEXT; // delay before next update
			break;

		case LIFE_KAPUT:
			if (saucersMoveX == 2)
			{
				SI_SetupObject(&sigodSaucers, SAUCERS_XMIN, SAUCERS_YPOS, CH_SAUCER, LIFE_ALIVE);
			}
			else
			{
				SI_SetupObject(&sigodSaucers, SAUCERS_XMAX, SAUCERS_YPOS, CH_SAUCER, LIFE_ALIVE);
			}

			updateTimer[REFRESH_SAUCERS] = TIMER_NOW; // display immediately
			break;
		}
	}
}

ft_void_t SI_UpdateInvader()
{
	if (SI_TIMER_EXPIRED(REFRESH_INVADER))
	{
		ft_uint8_t index;

		// kill off any intruder which was hit the last time ...
		for (index = 0; index < INVADER_TOTL; index++)
		{
			if (sigodInvader[index].life == LIFE_TOAST)
			{
				sigodInvader[index].life = LIFE_KAPUT;
			}
		}

		// move next invader ...
		for (index = 0; index < INVADER_TOTL; index++)
		{
			// if last invader has been checked ...
			if (invaderCheck == INVADER_TOTL)
			{
				// if any invader has reached edge ...
				if (invaderEdges)
				{
					// reverse X direction and drop down a line  ...
					invaderMoveX = (invaderMoveX == INVADER_LEFT)? INVADER_RIGHT:INVADER_LEFT;
					invaderMoveY = INVADER_YSPD;
					// reset edge reached ...
					invaderEdges = FALSE;
				}
				else
				{
					invaderMoveY = 0;
				}

				// start checking from the first invader again ...
				invaderCheck = 0;
			}

			if (sigodInvader[invaderCheck].life == LIFE_SPAWN)
			{
				sigodInvader[invaderCheck].life = LIFE_ALIVE;
			}

			// update position of invaders accordingly ...
			if (sigodInvader[invaderCheck].life == LIFE_ALIVE)
			{
				if (invaderMoveX == INVADER_LEFT)
				{
					sigodInvader[invaderCheck].xcrd -= INVADER_XSPD;
				}
				else
				{
					sigodInvader[invaderCheck].xcrd += INVADER_XSPD;
				}

				sigodInvader[invaderCheck].ycrd += invaderMoveY;
				sigodInvader[invaderCheck].anim  = ++sigodInvader[invaderCheck].anim % FRAME_INVADER;

				// determine if direction needs to change ...
				if ((sigodInvader[invaderCheck].xcrd <= INVADER_XMIN) || (sigodInvader[invaderCheck].xcrd >= INVADER_XMAX))
				{
					invaderEdges = TRUE;
				}
				else if (sigodInvader[invaderCheck].ycrd >= INVADER_YMAX)
				{
					invaderLands = TRUE;
				}

				// next invader to check (next time) ...
				invaderCheck++;
				break;
			}
			else
			{
				// next invader to check ...
				invaderCheck++;
			}
		}

		// set delay before next update ...
		updateTimer[REFRESH_INVADER] = TIMER_INVADER;
	}
}

ft_void_t SI_UpdatePlayers()
{
	if (SI_TIMER_EXPIRED(REFRESH_PLAYERS))
	{
		// set delay before next update ...
		updateTimer[REFRESH_PLAYERS] = TIMER_PLAYERS;

		switch (sigodPlayers.life)
		{
		case LIFE_ALIVE:
			// update position ...
			sigodPlayers.xcrd += playersMoveX;

			// keep within limits ...
			if (sigodPlayers.xcrd <= PLAYERS_XMIN)
			{
				sigodPlayers.xcrd = PLAYERS_XMIN;
			}
			else if (sigodPlayers.xcrd >= PLAYERS_XMAX)
			{
				sigodPlayers.xcrd = PLAYERS_XMAX;
			}

			playersMoveX = 0;
			break;

		case LIFE_TOAST:
			sigodPlayers.life = LIFE_KAPUT;
			break;

		case LIFE_KAPUT:
			break;

		case LIFE_SPAWN:
			if (++playersSpawn == 9)
			{
				sigodPlayers.life = LIFE_ALIVE;
			}
			else
			{
				updateTimer[REFRESH_PLAYERS] = TIMER_SPAWN;
			}
			break;
		}
	}
}

ft_void_t SI_UpdateBlaster()
{
	if (SI_TIMER_EXPIRED(REFRESH_BLASTER))
	{
		switch (sigodBlaster.life)
		{
		case LIFE_ALIVE:
			// update position and frame ...
			sigodBlaster.ycrd -= BLASTER_YSPD;
			sigodBlaster.anim  = ++sigodBlaster.anim % FRAME_BLASTER;

			if (sigodBlaster.ycrd <= BULLETS_YMIN)
			{
				sigodBlaster.ycrd = BULLETS_YMIN;
				sigodBlaster.life = LIFE_TOAST; // kill off bullet if it has reached top of display
			}
			else
			{
				sigodBlaster.life = SI_HitsOnTumbler(sigodBlaster.xcrd + 1, sigodBlaster.ycrd);
				if (sigodBlaster.life == LIFE_ALIVE)
				{
					sigodBlaster.life = SI_HitsOnZigZags(sigodBlaster.xcrd + 1, sigodBlaster.ycrd);
					if (sigodBlaster.life == LIFE_ALIVE)
					{
						sigodBlaster.life = SI_HitsOnStinger(sigodBlaster.xcrd + 1, sigodBlaster.ycrd);
						if (sigodBlaster.life == LIFE_ALIVE)
						{
							sigodBlaster.life = SI_HitsOnShields(sigodBlaster.xcrd + 1, sigodBlaster.ycrd);
							if (sigodBlaster.life == LIFE_ALIVE)
							{
								sigodBlaster.life = SI_HitsOnInvader(sigodBlaster.xcrd + 1, sigodBlaster.ycrd);
								if (sigodBlaster.life == LIFE_ALIVE)
								{
									sigodBlaster.life = SI_HitsOnSaucer(sigodBlaster.xcrd + 1, sigodBlaster.ycrd);
								}
							}
						}
					}
				}
			}

			updateTimer[REFRESH_BLASTER] = (sigodBlaster.life == LIFE_ALIVE)? TIMER_BLASTER : TIMER_EXPLODE;
			break;

		case LIFE_TOAST:
			sigodBlaster.life = LIFE_KAPUT;
			break;
		}
	}
}

ft_uint16_t SI_MissileTimer()
{
	// The higher the current score, the more frequent the missiles are dropped ...
	if      (scoreCurrent <  20) { return TIMER_MISSILE_UPTO_0200; }
	else if (scoreCurrent < 100) { return TIMER_MISSILE_UPTO_1000; }
	else if (scoreCurrent < 200) { return TIMER_MISSILE_UPTO_2000; }
	else if (scoreCurrent < 300) { return TIMER_MISSILE_UPTO_3000; }

	return TIMER_MISSILE_OVER_3000;
}

ft_void_t SI_UpdateMissile(ft_uint8_t type)
{
	// Handling of invader missiles (TUMBLER, ZIGZAGS and STINGER) ...
	ft_uint8_t invader = 255;
	ft_uint8_t missile = 0;

	if (SI_TIMER_EXPIRED(type))
	{
		switch (sigodMissile[type].life)
		{
		case LIFE_ALIVE:
			// update position and frame ...
			sigodMissile[type].ycrd += BULLETS_YSPD;
			sigodMissile[type].anim  = ++sigodMissile[type].anim % FRAME_MISSILE;

			if (sigodMissile[type].ycrd >= BULLETS_YMAX)
			{
				sigodMissile[type].ycrd = BULLETS_YMAX;
				sigodMissile[type].life = LIFE_TOAST; // kill off missile if it has reached bottom of display
			}
			else
			{
				sigodMissile[type].life = SI_HitsOnShields(sigodMissile[type].xcrd + 1, sigodMissile[type].ycrd + 8);
				if (sigodMissile[type].life == LIFE_ALIVE)
				{
					sigodMissile[type].life = SI_HitsOnPlayers(sigodMissile[type].xcrd + 1, sigodMissile[type].ycrd + 8);
				}
			}

			updateTimer[type] = (sigodMissile[type].life == LIFE_ALIVE)? TIMER_MISSILE_DROPPING : TIMER_EXPLODE;
			break;

		case LIFE_TOAST:
			sigodMissile[type].life = LIFE_KAPUT;
			// delay before next update ...
			updateTimer[type] = TIMER_MISSILE_DROPPING;
			break;

		case LIFE_KAPUT:
			if (SI_TIMER_EXPIRED(REFRESH_MISSILE))
			{
				sigodMissile[type].life = LIFE_SPAWN;
				// delay before next update ...
				updateTimer[type] = TIMER_NOW;
				// delay before next missile can be fired ...
				updateTimer[REFRESH_MISSILE] = SI_MissileTimer();
			}
			break;

		case LIFE_SPAWN:
			switch (type)
			{
			case REFRESH_TUMBLER:
				// select next column to drop TUMBLER missile in (handle wrap around as well) ...
				if (dropsTumbler[++indexTumbler] == 0)
				{
					indexTumbler = 0;
				}

				// find the index of the first invader alive in the column specified (if any) ...
				invader = InvaderColumn(dropsTumbler[indexTumbler]);
				missile = CH_TUMBLER;
				break;

			case REFRESH_ZIGZAGS:
				// select next column to drop ZIGZAGS missile in (handle wrap around as well) ...
				if (dropsZigZags[++indexZigZags] == 0)
				{
					indexZigZags = 0;
				}

				// find the index of the first invader alive in the column specified (if any) ...
				invader = InvaderColumn(dropsZigZags[indexZigZags]);
				missile = CH_ZIGZAGS;
				break;

			case REFRESH_STINGER:
				// find the index of the first invader alive vertically above the player (if any) ...
				invader = InvaderNearest();
				missile = CH_STINGER;
				break;
			}

			if (invader != 255)
			{
				// invader selected to drop missile ...
				SI_SetupObject(&sigodMissile[type], sigodInvader[invader].xcrd + 4, sigodInvader[invader].ycrd + 16, missile, LIFE_ALIVE);
			}

			updateTimer[type] = TIMER_MISSILE_DROPPING;
			break;
		}
	}
}

// =================================================================================================================
// SPACED-INVADERS - Add objects to display list
// =================================================================================================================
ft_void_t SI_RenderText(ft_uint16_t xcrd, ft_uint16_t ycrd, ft_uint8_t * text)
{
	// This version of text display requires 1 command per character (a shed load for the start screen), however
	// it allows us to display font 20 as a mono-spaced font (mind you some letters are a bit out of position).
	xcrd -= 1; // need to shift 1 pixel left to position character on 'x' correctly
	ycrd -= 3; // need to shift 3 pixels up  to position character on 'y' correctly

	// add text to display list ...
	do
	{
		if (*text == 'I')
		{
			// need to centre the letter 'I' ...
			SI_VertexII(xcrd+2, ycrd, 20, *text);
		}
		else
		{
			SI_VertexII(xcrd  , ycrd, 20, *text);
		}
		xcrd += 8;
	} while (*++text != 0);
}

ft_void_t SI_RenderExplode1(ft_uint8_t xcrd, ft_uint8_t ycrd)
{
	SI_VertexII(xcrd-2, ycrd, BH_PLAYERSAUCER, CH_EXPLODEX); // invader explosion
}

ft_void_t SI_RenderExplode2(ft_uint8_t xcrd, ft_uint8_t ycrd)
{
	SI_VertexII(xcrd-4, ycrd, BH_INVADERSEXPLODE, CH_EXPLODES); // other explosion
}

ft_void_t SI_InsertScore(ft_uint16_t value, ft_uint8_t xchar)
{
    ft_uint8_t  digit;
    ft_uint8_t  index;
    ft_uint16_t power[5] = {10000, 1000, 100, 10, 1};

    for (index = 0; index < 5; index++)
    {
        digit = 48;
        while (value >= power[index])
        {
            digit++;
            value -= power[index];
        }

        // insert digit
        displayScore[xchar++] = digit;
    }
}

ft_void_t SI_RenderInvader()
{
	ft_uint8_t index;
	ft_uint8_t lastcolour = 255;
	ft_uint8_t nextcolour;

	// render all invaders according to life state ...
	for (index = 0; index < INVADER_TOTL; index++)
	{
		// only add colour command if colour has changed (minimises commands to send)
		nextcolour = (sigodInvader[index].ycrd - INVADER_YMIN) / 8;

		if (nextcolour > 23)
		{
			nextcolour = 23;
		}

		if (lastcolour != nextcolour)
		{
			lastcolour = nextcolour;
			FT800_CmdBuffer_AddCmd( colours[nextcolour] );
		}

		switch (sigodInvader[index].life)
		{
		case LIFE_ALIVE:
			// add appropriate invader sprite to display list (depending on row and animation state) ...
			SI_VertexII(sigodInvader[index].xcrd, sigodInvader[index].ycrd, BH_INVADERSEXPLODE, sigodInvader[index].type + sigodInvader[index].anim);
			break;

		case LIFE_TOAST:
			// add appropriate toastiness graphic to display list ...
			SI_RenderExplode1(sigodInvader[index].xcrd, sigodInvader[index].ycrd);
			break;
		}
	}
}

ft_void_t SI_RenderShields()
{
	ft_uint8_t index;

	FT800_CmdBuffer_AddCmd( COLOR_RGB(0, 255, 0));

	for (index = 0; index < SHIELDS_TOTL; index++)
	{
		if (sigodShields[index].life > LIFE_TOAST)
		{
			// add appropriate shield graphic to display list ...
			SI_VertexII(sigodShields[index].xcrd, sigodShields[index].ycrd, BH_SHIELDS, sigodShields[index].type + (LIFE_SHIELD - sigodShields[index].life));
		}
	}
}

ft_void_t SI_RenderPlayers()
{
	ft_uint8_t lives = 0;

	// render player accordingly ...
	switch (sigodPlayers.life)
	{
	case LIFE_ALIVE:
		FT800_CmdBuffer_AddCmd(COLOR_RGB(  0,255,  0));
		if (sigodBlaster.life == LIFE_KAPUT)
		{
			SI_VertexII(sigodPlayers.xcrd, sigodPlayers.ycrd, BH_PLAYERSAUCER, CH_PLAYER + 0);
		}
		else
		{
			SI_VertexII(sigodPlayers.xcrd, sigodPlayers.ycrd, BH_PLAYERSAUCER, CH_PLAYER + 1);
		}
		break;

	case LIFE_TOAST:
		FT800_CmdBuffer_AddCmd(COLOR_RGB(255,  0,  0));
		SI_RenderExplode1(sigodPlayers.xcrd, sigodPlayers.ycrd);
		break;

	case LIFE_SPAWN:
		// only display player every 2nd timeout (makes it flash on/off) ...
		if ((playersSpawn % 2) == 0)
		{
			SI_VertexII(sigodPlayers.xcrd, sigodPlayers.ycrd, BH_PLAYERSAUCER, CH_PLAYER + 0);
		}
		break;
	}

	for (lives = 1; lives < playersLives; lives++)
	{
		SI_VertexII((lives * 20), 240, BH_PLAYERSAUCER, CH_PLAYER);
	}
}

ft_void_t SI_RenderSaucers()
{
	// render saucer accordingly ...
	switch (sigodSaucers.life)
	{
	case LIFE_ALIVE:
		FT800_CmdBuffer_AddCmd( COLOR_RGB(255, 0, 0) );
		SI_VertexII(sigodSaucers.xcrd, sigodSaucers.ycrd, BH_PLAYERSAUCER, CH_SAUCER + sigodSaucers.anim);
		break;

	case LIFE_TOAST:
		FT800_CmdBuffer_AddCmd( COLOR_RGB(255, 0, 0) );
		SI_RenderText(sigodSaucers.xcrd, sigodSaucers.ycrd, scoreSaucer);
		break;
	}
}

ft_void_t SI_RenderMissile(siObject_t * missile)
{
	// render bullet accordingly ...
	switch (missile->life)
	{
	case LIFE_ALIVE:
		SI_VertexII(missile->xcrd, missile->ycrd, BH_BULLETS, missile->type + missile->anim);
		break;

	case LIFE_TOAST:
		SI_RenderExplode2(missile->xcrd, missile->ycrd);
		break;
	}
}

ft_void_t SI_RenderScoring()
{
	// update high score if necessary ...
	if (scoreCurrent > scoreHighest)
	{
		scoreHighest = scoreCurrent;
	}

	FT800_CmdBuffer_AddCmd(COLOR_RGB(255, 255, 255));
	SI_InsertScore(scoreCurrent, 0);
	SI_InsertScore(scoreHighest, 9);
	SI_RenderText(siTextList[0].xoff*8, siTextList[0].yoff*8,siTextList[0].text);
	SI_RenderText(siTextList[1].xoff*8, siTextList[1].yoff*8,siTextList[1].text);
	SI_RenderText(siTextList[9].xoff*8, siTextList[9].yoff*8,siTextList[9].text);
}

ft_void_t SI_RenderBackground()
{
	ft_uint16_t size = 400;
	ft_uint8_t  fade;

	// Render sky ...
	FT800_CmdBuffer_AddCmd(CMD_GRADIENT);
	FT800_CmdBuffer_AddCmd((ft_uint32_t) 0);
	FT800_CmdBuffer_AddCmd(COLOR_RGB(  7, 22, 35));
	FT800_CmdBuffer_AddCmd((ft_uint32_t) DISPLAY_YPIX << 16);
	FT800_CmdBuffer_AddCmd(COLOR_RGB(  0, 46, 51));
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_BITMAPS);

	// Render moon ... yes, I know ... it's not very moon like (needs some craters)!
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_POINTS);
	// Add some alpha blending ...
	FT800_CmdBuffer_AddCmd(COLOR_A(128));

	for (fade = 0; fade < 16; fade++)
	{
		FT800_CmdBuffer_AddCmd(POINT_SIZE(size*16));
		FT800_CmdBuffer_AddCmd(COLOR_RGB(70 + fade * 2, 40 + fade * 2, fade * 2));
		FT800_CmdBuffer_AddCmd(VERTEX2F((DISPLAY_XOFF + 112) << 4, (DISPLAY_YOFF + 580 + fade * 8) << 4));
		size += 3;
	}
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_POINTS);

	// Render rectangle outlining actual play area (aiming for a raised panel look) ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_LINESTRIP);
	FT800_CmdBuffer_AddCmd(COLOR_RGB( 96, 96, 96));
	FT800_CmdBuffer_AddCmd(LINE_WIDTH(6));
	FT800_CmdBuffer_AddCmd(VERTEX2F((DISPLAY_XOFF               ) << 4, (DISPLAY_YOFF + DISPLAY_YPIX) << 4));
	FT800_CmdBuffer_AddCmd(VERTEX2F((DISPLAY_XOFF               ) << 4, (DISPLAY_YOFF               ) << 4));
	FT800_CmdBuffer_AddCmd(VERTEX2F((DISPLAY_XOFF + DISPLAY_XPIX) << 4, (DISPLAY_YOFF               ) << 4));
	FT800_CmdBuffer_AddCmd(COLOR_RGB(  0,  0,  0));
	FT800_CmdBuffer_AddCmd(VERTEX2F((DISPLAY_XOFF + DISPLAY_XPIX) << 4, (DISPLAY_YOFF + DISPLAY_YPIX) << 4));
	FT800_CmdBuffer_AddCmd(VERTEX2F((DISPLAY_XOFF               ) << 4, (DISPLAY_YOFF + DISPLAY_YPIX) << 4));
	// Reset to fully opaque ...
	FT800_CmdBuffer_AddCmd(COLOR_A(255));
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_LINESTRIP);

	// Render line towards bottom of display ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_BITMAPS);
	FT800_CmdBuffer_AddCmd(COLOR_RGB(  0,255,  0));
	SI_VertexII(0,238, BH_LONGLINE, CH_LONGLINE);
}

// =================================================================================================================
// SPACED-INVADERS - Get bitmap data and store on FT800
// =================================================================================================================
ft_int32_t SI_StoreBitmaps(ft_int32_t storeat, SAMAPP_Bitmap_header_t * bitmap_info, ft_uint8_t * bitmap_data)
{
	// calculate storage requirements for bitmap ...
	ft_int32_t bitmap_datasize = bitmap_info->Height * bitmap_info->Stride;

	// copy raw data for bitmap into memory ...
	Ft_Gpu_Hal_WrMemFromFlash(phost, storeat, bitmap_data, bitmap_datasize);

	// return address to store following data at...
	storeat += bitmap_datasize;
	return (storeat);
}

// =================================================================================================================
// SPACED-INVADERS - Insert bitmap handles etc into display list
// =================================================================================================================
ft_void_t SI_DlistBitmaps(ft_uint8_t handle, ft_int32_t storeat, SAMAPP_Bitmap_header_t * bitmap_info, ft_uint8_t cells)
{
	// add bitmap handle to display list ...
	FT800_CmdBuffer_AddCmd( BITMAP_HANDLE(handle));
	FT800_CmdBuffer_AddCmd( BITMAP_SOURCE(storeat));
	FT800_CmdBuffer_AddCmd( BITMAP_SIZE(NEAREST, BORDER, BORDER, bitmap_info->Width, bitmap_info->Height));
	FT800_CmdBuffer_AddCmd( BITMAP_LAYOUT(bitmap_info->Format, bitmap_info->Stride, bitmap_info->Height / cells));
}

ft_void_t SI_DisplayListStartNew()
{
	// Start of new display list ...
	FT800_CmdBuffer_AddCmd(CMD_DLSTART);

	// Start of BITMAPS section ...
	FT800_CmdBuffer_AddCmd(SECTION_BEGIN_BITMAPS);
}

ft_void_t SI_DisplayListTransmit()
{
	// End of BITMAPS section ...
	FT800_CmdBuffer_AddCmd(SECTION_CLOSE_BITMAPS);

	// end of display list ...
	FT800_CmdBuffer_AddCmd(DISPLAY());

	// Swap between frames
	FT800_CmdBuffer_AddCmd(CMD_SWAP);

	// insert next line to see how many commands have been sent ...
	//SI_InsertScore( (Ft_CmdBuffer_Index / 4), 19);

	// Download the commands into fifo
	FT800_CmdBuffer_Send();
}

ft_void_t SI_CheckButtons()
{
	if (SI_TIMER_EXPIRED(REFRESH_BUTTONS))
	{
		checkButtons = ((~P0) >> 1) & BUTTON_MASK;
		playersMoveX = 0;

		if (checkButtons & BUTTON_LEFT)
		{
			playersMoveX = -PLAYERS_XSPD;
		}

		if (checkButtons & BUTTON_RIGHT)
		{
			playersMoveX = +PLAYERS_XSPD;
		}

		if (checkButtons & BUTTON_FIRE)
		{
			// only allow one shot at a time and when player is alive ...
			if ((sigodPlayers.life == LIFE_ALIVE) && (sigodBlaster.life == LIFE_KAPUT))
			{
				// fire blaster ...
				updateTimer[REFRESH_BLASTER] = TIMER_NOW;
				SI_SetupObject(&sigodBlaster, sigodPlayers.xcrd + 6, PLAYERS_YPOS, CH_BLASTER, LIFE_ALIVE);

				// saucer score related to number of blaster shots fired ...
				blasterCount = ++blasterCount % 15;

				// saucer direction reverses every 15th shot of blaster ...
				if (blasterCount == 0)
				{
					if (sigodSaucers.life != LIFE_ALIVE)
					{
						saucersMoveX = -saucersMoveX;
					}
				}
			}
		}

		updateTimer[REFRESH_BUTTONS] = TIMER_BUTTONS;
	}
}

ft_void_t SI_MsWait(ft_uint16_t wait)
{
	ms_timer = wait;
	while (ms_timer);
}

ft_void_t SI_WaitForButton(ft_uint8_t button)
{
	do
	{
		SI_MsWait(10);
		checkButtons = ((~P0) >> 1) & BUTTON_MASK;
	} while (checkButtons != button);
}

ft_void_t SI_WaitForStart()
{
	ft_uint8_t index;

	// start new display list ...
	SI_DisplayListStartNew();

	// add bitmap handles to display list ...
	SI_DlistBitmaps(BH_INVADERSEXPLODE, ramInvadersExplode, InvadersExplodeImageInfo, 12);
	SI_DlistBitmaps(BH_SHIELDS,         ramShields,         ShieldsImageInfo,         20);
	SI_DlistBitmaps(BH_BULLETS,         ramMissile,         MissileImageInfo,         16);
	SI_DlistBitmaps(BH_PLAYERSAUCER,    ramPlayerSaucer,    PlayerSaucerImageInfo,     5);
	SI_DlistBitmaps(BH_LONGLINE,        ramLongLine,        LongLineImageInfo,         1);

	// add background and startup text ...
	SI_RenderBackground();
	SI_RenderScoring();
	for (index = 2; index < 10; index++)
	{
		SI_RenderText(siTextList[index].xoff*8, siTextList[index].yoff*8,siTextList[index].text);
	}

	// add invader images to score table ...
	SI_VertexII(8*8-2, 8*17, BH_PLAYERSAUCER,    CH_SAUCER);
	SI_VertexII(8*8  , 8*19, BH_INVADERSEXPLODE, CH_INVADER1+1);
	SI_VertexII(8*8  , 8*21, BH_INVADERSEXPLODE, CH_INVADER3  );
	SI_VertexII(8*8  , 8*23, BH_INVADERSEXPLODE, CH_INVADER5+1);

	// finish display list and send to FT800 ...
	SI_DisplayListTransmit();

	// wait until all buttons are released ...
	SI_WaitForButton(BUTTON_NONE);

	// wait until 'START' button is pressed and released ...
	SI_WaitForButton(BUTTON_START);
	SI_WaitForButton(BUTTON_NONE);
}

// =================================================================================================================
// SPACED-INVADERS - Main control
// =================================================================================================================
ft_void_t SpacedInvaders()
{
	ramInvadersExplode = RAM_G;
	ramShields         = SI_StoreBitmaps(ramInvadersExplode, InvadersExplodeImageInfo, InvadersExplodeImageData);
	ramMissile         = SI_StoreBitmaps(ramShields,         ShieldsImageInfo,         ShieldsImageData);
	ramPlayerSaucer    = SI_StoreBitmaps(ramMissile,         MissileImageInfo,         MissileImageData);
	ramLongLine        = SI_StoreBitmaps(ramPlayerSaucer,    PlayerSaucerImageInfo,    PlayerSaucerImageData);
	ramNextBit         = SI_StoreBitmaps(ramLongLine,        LongLineImageInfo,        LongLineImageData);

    // Initialise the timer to provide millisecond countdown timer(s) ...
    ms_timer_initialise();

	do
	{
		// wait for START switch to be pressed ...
		SI_WaitForStart();

		// start game (reset and start first wave) ...
		SI_BeginGame(GAME_START_NEW);

		do // play game ...
		{
			do // play wave ...
			{
				SI_CheckButtons();
				SI_UpdateSaucers();
				SI_UpdateInvader();
				SI_UpdatePlayers();
				SI_UpdateBlaster();
				SI_UpdateMissile(REFRESH_TUMBLER);
				SI_UpdateMissile(REFRESH_ZIGZAGS);
				SI_UpdateMissile(REFRESH_STINGER);

				if (SI_TIMER_EXPIRED(REFRESH_DISPLAY))
				{
					// start new display list ...
					SI_DisplayListStartNew();

					// add all game objects to display list ...
					SI_RenderBackground();
					SI_RenderPlayers();
					SI_RenderShields();
					SI_RenderInvader();
					SI_RenderSaucers();
					SI_RenderScoring();
					SI_RenderMissile(&sigodBlaster);
					SI_RenderMissile(&sigodMissile[REFRESH_TUMBLER]);
					SI_RenderMissile(&sigodMissile[REFRESH_ZIGZAGS]);
					SI_RenderMissile(&sigodMissile[REFRESH_STINGER]);

					// finish display list and send to FT800 ...
					SI_DisplayListTransmit();

					// reset screen display update timer ...
					updateTimer[REFRESH_DISPLAY] = TIMER_DISPLAY;
				}

				// deal with player being hit by an invader missile ...
				if  (sigodPlayers.life == LIFE_KAPUT)
				{
					// continue with wave if player has some lives left ...
					if (playersLives > 1)
					{
						playersLives--;
						playersSpawn = 0;
						SI_SetupObject(&sigodPlayers, PLAYERS_XPOS, PLAYERS_YPOS, CH_PLAYER, LIFE_SPAWN);
						updateTimer[REFRESH_PLAYERS] = TIMER_NOW;
					}
				}
			} while ((invaderAlive != 0) && (invaderLands == FALSE) && (sigodPlayers.life != LIFE_KAPUT));

			if ((invaderLands == FALSE) && (sigodPlayers.life == LIFE_ALIVE))
			{
				// reset invaders for next wave ...
				SI_BeginGame(GAME_NEXT_WAVE);
			}

		} while ((invaderLands == FALSE) && (sigodPlayers.life == LIFE_ALIVE));

	} while (1);
}

// =================================================================================================================
// SPACED-INVADERS - END OF CODE
// =================================================================================================================